/*
** rts/command.c
**
** Viper RTS Command functions
**
** Copyright (c) 1998 Peter Eriksson <pen@lysator.liu.se>
*/


#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <time.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>

#include "rts/rts.h"
#include "rts/scsi.h"

int
rts_rebuild_disk_drive(RTS *rp,
		       int timeout,
		       int raid_set,
		       int channel,
		       int scsi_id)
{
    int code;
    struct rts_rebuild_disk_drive rddb;

    
    rddb.timeout = timeout;
    rddb.raid_set = raid_set;
    rddb.channel = channel;
    rddb.scsi_id = scsi_id;
    
    code = rts_command(rp, RTS_REBUILD_DISK_DRIVE, &rddb, sizeof(rddb));
    return code;
}


int
rts_create_raid_set(RTS *rp,
		    struct rts_create_raid_set *crsp)
{
    int code;

    
    code = rts_command(rp, RTS_CREATE_RAID_SET, crsp, sizeof(*crsp));
    return code;
}


int
rts_add_spare(RTS *rp,
	      int timeout,
	      int spare_designation,
	      int channel,
	      int scsi_id)
{
    int code;
    struct rts_add_spare asb;


    asb.timeout = timeout;
    asb.spare_designation = spare_designation;
    asb.channel = channel;
    asb.scsi_id = scsi_id;
    
    code = rts_command(rp, RTS_ADD_SPARE, &asb, sizeof(asb));
    return code;
}


int
rts_change_password(RTS *rp,
		    int password_type,
		    const char *new_password)
{
    int code, len;
    struct rts_change_password cpb;

    len = strlen(new_password);
    if (len > 9)
    {
	rp->unix_error_code = EINVAL;
	return -1;
    }

    memset(&cpb, 0, sizeof(cpb));
    cpb.password_type = password_type;
    strcpy(cpb.new_password, new_password);
    
    code = rts_command(rp, RTS_CHANGE_PASSWORD, &cpb, sizeof(cpb));
    return code;
}


int
rts_scan_drive(RTS *rp,
	       int timeout,
	       int channel,
	       int scsi_id)
{
    int code;
    struct rts_scan_drive sdb;


    sdb.timeout = timeout;
    sdb.channel = channel;
    sdb.scsi_id = scsi_id;
    
    code = rts_command(rp, RTS_SCAN_DRIVE, &sdb, sizeof(sdb));
    return code;
}


int
rts_flush_event_log(RTS *rp)
{
    int code;

    
    code = rts_command(rp, RTS_FLUSH_EVENT_LOG, NULL, 0);
    return code;
}



int
rts_begin_shutdown(RTS *rp)
{
    int code;

    
    code = rts_command(rp, RTS_BEGIN_SHUTDOWN, NULL, 0);
    return code;
}


int
rts_complete_shutdown(RTS *rp,
		      int final_shutdown_action)
{
    int code;
    struct rts_complete_shutdown csb;

    
    csb.final_shutdown_action = final_shutdown_action;
    
    code = rts_command(rp, RTS_COMPLETE_SHUTDOWN, &csb, sizeof(csb));
    return code;
}


int
rts_drop_drive(RTS *rp,
	       int timeout,
	       int channel,
	       int scsi_id)
{
    int code;
    struct rts_drop_drive ddb;    


    ddb.timeout = timeout;
    ddb.channel = channel;
    ddb.scsi_id = scsi_id;
    
    code = rts_command(rp, RTS_DROP_DRIVE, &ddb, sizeof(ddb));
    return code;
}


int
rts_begin_session(RTS *rp,
		  int session_type)
{
    struct rts_message msg;
    unsigned long tag;
    int code;
    
    
    rts_clearerr(rp);
    tag = rts_init_message(&msg, RTS_COMMAND, RTS_BEGIN_SESSION,
			   rp->ownership_token);

    msg.body.command.begin_session.session_type = session_type;
    
    if (rts_lock(rp) < 0)
	return -1;
    
    code = scsi_command(rp->fd, SCMD_SDIAG, USCSI_WRITE,
			(void  *) &msg, 141, SCSI_TIMEOUT);
    if (code != 0)
    {
	rp->unix_error_code = errno;
	rts_unlock(rp);
	return -1;
    }

    memset(&msg, 0, sizeof(msg));
    
    code = scsi_command(rp->fd, SCMD_GDIAG, USCSI_READ, 
			    (void *) &msg, sizeof(msg), SCSI_TIMEOUT);

    if (code != 0)
    {
	rp->unix_error_code = errno;
	rts_unlock(rp);
	return -1;
    }

    rts_unlock(rp);
    
    if (rts_valid_acknowledge(&msg, tag) < 0)
    {
	rp->ack.s.status = 254;
	return -1;
    }

    rp->ack = msg.body.acknowledge;
    
    if (msg.body.acknowledge.s.status != 0)
	return msg.body.acknowledge.s.status;

    rp->ownership_token = msg.ownership_token;
    
    return 0;
}



int
rts_end_session(RTS *rp)
{
    int code;


    code = rts_command(rp, RTS_END_SESSION, NULL, 0);
    if (code != 0)
	return code;
    
    rp->ownership_token = 0;
    
    return 0;
}


int
rts_format_drive(RTS *rp,
		 int timeout,
		 int channel,
		 int scsi_id)
{
    int code;
    struct rts_format_drive fdb;

    fdb.timeout = timeout;
    fdb.channel = channel;
    fdb.scsi_id = scsi_id;
    
    code = rts_command(rp, RTS_FORMAT_DRIVE, &fdb, sizeof(fdb));
    return code;
}


int
rts_delete_raid_set(RTS *rp,
		    int timeout,
		    int raid_set)
{
    int code;
    struct rts_delete_raid_set drsb;

    drsb.timeout = timeout;
    drsb.raid_set = raid_set;
    
    code = rts_command(rp, RTS_DELETE_RAID_SET, &drsb, sizeof(drsb));
    return code;
}



int
rts_set_alarm_state(RTS *rp,
		    int alarm_state)
{
    int code;
    struct rts_set_alarm_state asb;

    
    asb.alarm_state = alarm_state;
    
    code = rts_command(rp, RTS_SET_ALARM_STATE, &asb, sizeof(asb));
    return code;
}


int
rts_switchover_controller(RTS *rp)
{
    int code;

    
    code = rts_command(rp, RTS_SWITCHOVER_CONTROLLER, NULL, 0);
    return code;
}


