/*
** rts/rdwr.h
**
** Viper RTS Read and Write functions
**
** Copyright (c) 1998 Peter Eriksson <pen@lysator.liu.se>
*/

#ifndef VIPER_RTS_RDWR_H
#define VIPER_RTS_RDWR_H


/* RTS Read and Write Messages */
#define RTS_HOST_PARAMETERS			0
#define RTS_SYSTEM_PARAMETERS			2
#define RTS_VENDOR_PARAMETERS			3
#define RTS_MODE_PARAMETERS			5
#define RTS_DATE_AND_TIME			6
#define RTS_DISK_PARAMETERS			7
#define RTS_REDUNDANT_CONTROLLER_PARAMETERS	8
#define RTS_CACHE_PARAMETERS			9

/* More RTS Read Messages */
#define RTS_STATIC_SYSTEM_INFORMATION		130
#define RTS_DYNAMIC_SYSTEM_INFORMATION		131
#define RTS_REDUNDANCY_GROUP_INFORMATION	132
#define RTS_ENVIRONMENTAL_INFORMATION		133
#define RTS_DISK_DRIVE_INFORMATION   		134
#define RTS_EVENT_LOG_INFORMATION    		135
#define RTS_EVENT_INFORMATION       		136
#define RTS_CREATE_INFORMATION       		137
#define RTS_REBUILD_INFORMATION      		138
#define RTS_SHUTDOWN_INFORMATION     		139
#define RTS_RAID_SET_INFORMATION     		140
#define RTS_REDUNDANT_CONTROLLER_INFORMATION	141
#define RTS_IO_STATISTICAL_INFORMATION		142
#define RTS_SESSION_INFORMATION      		143


/*
** RTS Write and Response Message Structures
*/

/* Host Parameters */
struct rts_host_parameters
{
    Byte channel;
    Byte scsi_id;
    Byte tag_queuing;
    Byte sync_mode;
    Byte sync_rate;
    Byte bus_width;
    Byte termination;
    Byte host_lun_map[32];
    Byte inquiry_peripheral;
};

/* System Parameters */
struct rts_system_parameters
{
    Byte password_checking;
    Byte rebuild_rate;
    Byte create_rate;
    Byte ups_installed;
    Byte ch1_module_type;
    Byte ch2_module_type;
    Byte ch3_module_type;
    Byte read_ahead_enabled;
    Byte read_ahead_limit;
    Byte validation_delay;
    Byte twelve_volt_sensor_enabled;
    Byte host_wce_enabled;
    Byte drive_command_timeout;
    Byte reset_propagation;
    Byte environmental_alerts;
    Byte spinup_staggering;
};

/* Vendor Parameters */
struct rts_vendor_parameters
{
    Byte system_display_vendor[16];
    Byte system_display_model[16];
    Byte inquiry_response_vendor[8];
    Byte inquiry_response_model[16];    /* Protocol version 1   - 68 bytes */
#if RTS_VERSION >= 2
    Byte percent_write_back_cache;	/* Protocol version 2   - 69 bytes */
    Byte enclosure_severe_level_temp;
#if RTS_VERSION >= 5
    Byte enclosure_warning_level_temp;  /* Protocol version 5-6 - 71 bytes */
#if RTS_VERSION >= 7
    Byte extended_acknowledge_mode;	/* Protocol version 7-  - 72 bytes */
#endif
#endif
#endif
};

/* Mode Parameters */
struct rts_mode_parameters
{
    Byte channel;
    Byte read_write_error_recovery_page[12];
    Byte disconnect_reconnect_page[16];
    Byte format_device_page[24];
    Byte rigid_disk_drive_geometry_page[24];
    Byte verify_error_recovery_page[12];
    Byte caching_page[12];
    Byte peripheral_device_page[8];
    Byte control_mode_page[8];
    Byte medium_types_supported_page[8];
    Byte notch_and_partition_page[24];
};

/* Disk Parameters */
struct rts_disk_parameters
{
    Byte disk_channel_termination[10];
    Byte disk_channel_transfer_mode[10];
};

/* Redundant Controller Parameters */
struct rts_redundant_controller_parameters
{
    Byte controller_name[10];
    Byte host_channel_status[4];
};

/* Cache Parameters */
struct rts_cache_parameters
{
    Quad host0_lun_map;
    Quad host1_lun_map;
    Quad host2_lun_map;
    Quad host3_lun_map;
};

union rts_write_message
{
    struct rts_host_parameters		host_parameters;
    struct rts_system_parameters	system_parameters;
    struct rts_vendor_parameters	vendor_parameters;
    struct rts_mode_parameters		mode_parameters;
    struct rts_date_and_time 		date_and_time;
    struct rts_disk_parameters		disk_parameters;
    struct rts_redundant_controller_parameters redundant_controller_parameters;
    struct rts_cache_parameters		cache_parameters;
};


/*
** RTS Read Message Structures
*/
struct rts_read_host_parameters
{
    Byte channel;
};

struct rts_read_mode_parameters
{
    Byte channel;
};

struct rts_read_redundancy_group_information
{
    Byte redundancy_group;
};

struct rts_read_disk_drive_information
{
    Byte channel;
    Byte scsi_id;
};

struct rts_read_event_information
{
    Word event_sequence_number;
};

struct rts_read_create_information
{
    Byte raid_set_number;
};

struct rts_read_rebuild_information
{
    Byte channel;
    Byte scsi_id;
};

struct rts_read_raid_set_information
{
    Byte raid_set;
};

struct rts_read_io_statistical_information
{
    Quad redundancy_groups;
    Byte sample_period;
};


union rts_read_message
{
    struct rts_read_host_parameters 		 host_parameters;
    struct rts_read_mode_parameters 		 mode_parameters;
    struct rts_read_redundancy_group_information redundancy_group_information;
    struct rts_read_disk_drive_information	 disk_drive_information;
    struct rts_read_event_information		 event_information;
    struct rts_read_create_information		 create_information;
    struct rts_read_rebuild_information		 rebuild_information;
    struct rts_read_raid_set_information	 raid_set_information;
    struct rts_read_io_statistical_information	 io_statistical_information;
};




/*
** RTS Response Messages
*/


/* Static System Information */
struct rts_static_system_information
{
    Byte firmware_version[9];
    Byte rts_version;
    Byte model_number;
    Byte processor_type;
    Byte processor_clock;
    Byte processor_memory;
    Word cache_size;
    Byte simm_0a_size;
    Byte simm_0b_size;
    Byte simm_1a_size;
    Byte simm_1b_size;
    Byte serial_number[9];
    Byte manufacture_month;
    Byte manufacture_date;
    Byte manufacture_year;
    Byte channel_status[10];
    Byte dma_vendor_id[10];
    Byte dram_0_vendor_id;
    Byte dram_1_vendor_id;
    Byte xor_vendor_id;
    Byte pic_vendor_id;
    Byte reserved_1;
    Byte reserved_2;
};


/* Dynamic System Information */
struct rts_dynamic_system_information
{
    Quad online_redundancy_groups;
    Quad online_raid_sets;
    Quad creating_raid_sets;
    Word channel_0_disk_drives;
    Word channel_1_disk_drives;
    Word channel_2_disk_drives;
    Word channel_3_disk_drives;
    Word channel_4_disk_drives;
    Word channel_5_disk_drives;
    Word channel_6_disk_drives;
    Word channel_7_disk_drives;
    Word channel_8_disk_drives;
    Word channel_9_disk_drives;
    Word channel_0_rebuilding_drives;
    Word channel_1_rebuilding_drives;
    Word channel_2_rebuilding_drives;
    Word channel_3_rebuilding_drives;
    Word channel_4_rebuilding_drives;
    Word channel_5_rebuilding_drives;
    Word channel_6_rebuilding_drives;
    Word channel_7_rebuilding_drives;
    Word channel_8_rebuilding_drives;
    Word channel_9_rebuilding_drives;
    Word channel_0_spare_drives;
    Word channel_1_spare_drives;
    Word channel_2_spare_drives;
    Word channel_3_spare_drives;
    Word channel_4_spare_drives;
    Word channel_5_spare_drives;
    Word channel_6_spare_drives;
    Word channel_7_spare_drives;
    Byte channel_8_spare_drives;
    Word channel_9_spare_drives;
    Byte remote_access_level;
};


/* Redundancy Group Information */
struct rts_redundancy_group_information
{
    Byte redundancy_group;
    Byte raid_set;
    Byte raid_status;
    Byte raid_level;
    Byte host_0_lun;
    Byte host_1_lun;
    Byte host_2_lun;
    Byte host_3_lun;
    struct rts_date_and_time time_stamp;
    Word chunk_size;
    Quad capacity;
    Byte number_of_members;
    struct
    {
	Byte channel;
	Byte scsi_id;
    } member_descriptor[111];
};


/* Environmental Information */
struct rts_environmental_information
{
    Word five_volt_level;
    Word twelve_volt_level;
    Word scsi_termination_volt_level;
    Word dram_0_volt_level;
    Word dram_1_volt_level;
    Word backup_batteri_volt_level;
    Byte board_temperature_level;
    Byte enclosure_temperature_level;
    Byte five_volt_status;
    Byte twelve_volt_status;
    Byte scsi_termination_volt_status;
    Byte dram_0_volt_status;
    Byte dram_1_volt_status;
    Byte backup_battery_volt_status;
    Byte board_temperature_status;
    Byte enclosure_temperature_status;
    Byte calibration_status;
    Byte ups_outputs;
    Byte ups_status;
};


/* Disk Drive Information */
struct rts_disk_drive_information
{
    Byte channel;
    Byte scsi_id;
    Byte status;
    Byte raid_set;
    Byte member;
    Quad capacity;
    Byte disk_vendor[8];
    Byte disk_model[16];
    Byte disk_revision[4];
};


/* Create Information */
struct rts_create_information
{
    Quad blocks_done;
    Quad total_blocks;
    struct rts_date_and_time time_stamp;
    struct rts_date_and_time current;
    Byte status;
    Byte raid_set_number;
    Byte error_status;
    Byte error_channel;
    Byte error_id;
    Byte port_status;
    Byte scsi_status;
    Byte request_sense[14];
};

/* Rebuild Information */
struct rts_rebuild_information
{
    Byte channel;
    Byte scsi_id;
    Byte member;
    Byte status;
    Quad blocks_done;
    Quad total_blocks;
    struct rts_date_and_time time_stamp;
    struct rts_date_and_time current;
    Byte raid_set;
    Byte error_status;
    Byte error_channel;
    Byte error_scsi_id;
    Byte port_status;
    Byte scsi_status;
    Byte request_sense[14];
};

/* Shutdown Information */
struct rts_shutdown_information
{
    Quad modified_block_count;
    Byte status;
};

/* RAID Set Information */
struct rts_raid_set_information
{
    Byte raid_set;
    Byte raid_status;
    Byte raid_level;
    Byte number_of_redundancy_groups;
    Word number_of_members;
    struct rts_date_and_time time_stamp;
    Byte data[231]; /* Complex structure - see manual */
};

/* Redundant Controller Information */
struct rts_redundant_controller_information
{
    Byte unit_a_name[10];
    Byte unit_b_name[10];
    Byte unit_a_scsi_id;
    Byte unit_b_scsi_id;
    Byte unit_a_active_hosts;
    Byte unit_b_active_hosts;
    Quad unit_a_active_raid_sets;
    Quad unit_b_active_raid_sets;
    Quad unit_a_active_redundancy_groups;
    Quad unit_b_active_redundancy_groups;
    Word flags;
    Byte status;
    Byte reason;
    union 
    {
	struct
	{
	    Byte type;
	    Byte channel;
	} timeout;
	struct
	{
	    Quad other_controllers_cache_size;
	    Quad this_controllers_cache_size;
	} cache_mismatch;
	struct
	{
	    Byte other_controllers_fw_version[4];
	    Byte this_controllers_fw_version[4];
	} firmware_version;
	struct
	{
	    Byte other_controllers_slot_status[10];
	    Byte this_controllers_slot_status[10];
	} hardware_mismatch;
	struct
	{
	    Byte channel;
	} hardware_error;
	struct
	{
	    Byte channel;
	} parity_error;
	struct
	{
	    Byte redundancy_group;
	} raid_set_error;
	struct
	{
	    Quad raid_set_bitmap;
	    Quad redundancy_group_bitmap;
	} host_map_error;
	struct
	{
	    Byte unexpected_opcode;
	} unexpected_message;
	struct
	{
	    Byte raid_set_number;
	} redundant_controller_set_broken;
	Byte dummy[22];
    } data;
};

/* I/O Statistical Information */
struct rts_io_statistical_information
{
    Quad redundancy_groups;
    Quad sample_count;
    struct {
	Quad number_of_reads;
	Quad number_of_writes;
	Quad total_blocks_read;
	Quad total_blocks_written;
	Quad cache_hits;
	Quad cache_misses;
    } group[1];
};

/* Session Information */
struct rts_session_information
{
    Byte user_information[64];
    Byte system_information[64];
    Byte session_established;
};


union rts_response_message
{
    struct rts_host_parameters		host_parameters;
    struct rts_system_parameters	system_parameters;
    struct rts_vendor_parameters	vendor_parameters;
    struct rts_mode_parameters		mode_parameters;
    struct rts_date_and_time 		date_and_time;
    struct rts_disk_parameters		disk_parameters;
    struct rts_redundant_controller_parameters redundant_controller_parameters;
    struct rts_cache_parameters		cache_parameters;

    struct rts_static_system_information 	static_system_information;
    struct rts_dynamic_system_information 	dynamic_system_information;
    struct rts_redundancy_group_information	redundancy_group_information;
    struct rts_environmental_information	environmental_information;
    struct rts_disk_drive_information		disk_drive_information;
    struct rts_event_log_information		event_log_information;
    struct rts_event_information		event_information;
    struct rts_create_information		create_information;
    struct rts_rebuild_information		rebuild_information;
    struct rts_shutdown_information		shutdown_information;
    struct rts_raid_set_information		raid_set_information;
    struct rts_redundant_controller_information redundant_controller_information;
    struct rts_io_statistical_information	io_statistical_information;
    struct rts_session_information		session_information;
    
};

/*
** RTS Read Message functions
*/

extern int
rts_read_host_parameters(RTS *rp,
			 int channel,
			 struct rts_host_parameters *hpp);

extern int
rts_read_date_and_time(RTS *rp, time_t *bt);

extern int
rts_read_static_system_information(RTS *rp,
				   struct rts_static_system_information *sp);

extern int
rts_read_dynamic_system_information(RTS *rp,
				    struct rts_dynamic_system_information *sp);

extern int
rts_read_redundancy_group_information(RTS *rp,
				      int redundancy_group,
				      struct rts_redundancy_group_information *sp);

extern int
rts_read_environmental_information(RTS *rp,
				    struct rts_environmental_information *sp);

extern int
rts_read_disk_drive_information(RTS *rp,
				int channel,
				int scsi_id,
				struct rts_disk_drive_information *sp);

extern int
rts_read_event_log_information(RTS *rp,
			       struct rts_event_log_information *sp);

extern int
rts_read_event_information(RTS *rp,
			   int event_sequence_number,
			   struct rts_event_information *sp);

extern int
rts_read_create_information(RTS *rp,
			    int raid_set_number,
			    struct rts_create_information *sp);

extern int
rts_read_rebuild_information(RTS *rp,
			     int channel,
			     int scsi_id,
			     struct rts_rebuild_information *sp);

extern int
rts_read_shutdown_information(RTS *rp,
			      struct rts_shutdown_information *sp);

extern int
rts_read_raid_set_information(RTS *rp,
			      int raid_set,
			      struct rts_raid_set_information *sp);

extern int
rts_read_redundant_controller_information(RTS *rp,
			struct rts_redundant_controller_information *sp);

extern int
rts_read_io_statistical_information(RTS *rp,
				    unsigned long redundancy_groups,
				    int sample_period,
				    struct rts_io_statistical_information *sp);

extern int
rts_read_session_information(RTS *rp,
			     struct rts_session_information *sp);




/*
** RTS Write Message functions
*/
extern int
rts_write_host_parameters(RTS *rp,
			  struct rts_host_parameters *hpp);

extern int
rts_write_system_parameters(RTS *rp,
			    struct rts_system_parameters *spp);

extern int
rts_write_vendor_parameters(RTS *rp,
			    struct rts_vendor_parameters *vpp);

extern int
rts_write_mode_parameters(RTS *rp,
			  struct rts_mode_parameters *mpp);

extern int
rts_write_date_and_time(RTS *rp,
			time_t bt);

extern int 
rts_write_disk_parameters(RTS *rp,
			  struct rts_disk_parameters *dpp);

extern int
rts_write_redundant_controller_parameters(RTS *rp,
			  struct rts_redundant_controller_parameters *rcpp);


extern int
rts_write_cache_parameters(RTS *rp,
			   struct rts_cache_parameters *cpp);


#endif
