/*
** rts.c
**
** Viper Remote Terminal Service functions
**
** Copyright (c) 1998 Peter Eriksson <pen@lysator.liu.se>
*/

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <time.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>

#include "rts/rts.h"
#include "rts/scsi.h"


#define MAX_LOCK_ATTEMPTS	5
#define PATH_LOCKFILE		"/var/spool/locks/.rts_lock"



static int debug = 0;

unsigned int
rts_q2i(Quad q)
{
    return ((unsigned int) (q.b[0]) +
	    ((unsigned int) (q.b[1])<<8) +
	    ((unsigned int) (q.b[2])<<16) +
	    ((unsigned int) (q.b[3])<<24));
}

unsigned int
rts_w2i(Word w)
{
    return ((unsigned int) (w.b[0]) +
	    ((unsigned int) (w.b[1])<<8));
}


Quad
rts_i2q(unsigned int i)
{
    Quad q;
    
    q.b[0] = i & 0xFF;
    q.b[1] = (i>>8) & 0xFF;
    q.b[2] = (i>>16) & 0xFF;
    q.b[3] = (i>>24) & 0xFF;

    return q;
}


Word
rts_i2w(unsigned int i)
{
    Word w;
    
    w.b[0] = i & 0xFF;
    w.b[1] = (i>>8) & 0xFF;

    return w;
}


struct rts_date_and_time *
rts_localtime(time_t t,
	      struct rts_date_and_time *rtp)
{
    static struct rts_date_and_time datb;
    struct tm tmb;


    if (rtp == NULL)
	rtp = &datb;
    
    if (localtime_r(&t, &tmb) == NULL)
	return NULL;
    
    rtp->year   = tmb.tm_year;
    rtp->month  = tmb.tm_mon+1;
    rtp->date   = tmb.tm_mday;
    rtp->hour   = tmb.tm_hour;
    rtp->minute = tmb.tm_min;
    rtp->second = tmb.tm_sec;
    
    return rtp;
}

time_t
rts_mktime(struct rts_date_and_time *rtp)
{
    struct tm o_tmb, tmb;
    time_t bt;

    time(&bt);
    localtime_r(&bt, &o_tmb);
    
    memset(&tmb, 0, sizeof(tmb));
    tmb.tm_year = rtp->year;
    tmb.tm_mon  = rtp->month - 1;
    tmb.tm_mday = rtp->date;
    tmb.tm_hour = rtp->hour;
    tmb.tm_min  = rtp->minute;
    tmb.tm_sec  = rtp->second;
    tmb.tm_isdst = o_tmb.tm_isdst;
    
    return mktime(&tmb);
}


void
rts_set_debug(int flag)
{
    debug = flag;
}


static unsigned int
make_message_tag(void)
{
    static unsigned int c = 0;

    if (c == 0)
	c = random();
    
    return c++;
}

unsigned long
rts_init_message(struct rts_message *msg,
		 int message_type,
		 int message_type_extender,
		 int ownership_token)
{
    unsigned long tag;

    
    memset(msg, 0, sizeof(*msg));
    msg->signature[0] = 'R';
    msg->signature[1] = 'T';
    msg->signature[2] = 'S';
    msg->signature[3] = ' ';
    msg->protocol_version = RTS_VERSION;
    msg->ownership_token = ownership_token;
    msg->message_type = message_type;
    msg->message_type_extender = message_type_extender;
    
    tag = make_message_tag();
    msg->message_tag = rts_i2q(tag);
    
    return tag;
}
    
int
rts_valid_response(struct rts_message *msg,
		   unsigned long tag,
		   int message_type_extender)
{
    if (memcmp(msg->signature, " STR", 4) != 0)
	return -1;
    
    if (rts_q2i(msg->message_tag) != tag)
	return -2;

    if (msg->message_type != RTS_RESPONSE)
	return -3;

    if (msg->message_type_extender != message_type_extender)
	return -4;

    return 0;
}


int
rts_valid_acknowledge(struct rts_message *msg,
		      unsigned long tag)
{
    if (memcmp(msg->signature, " STR", 4) != 0)
	return -1;
    
    if (rts_q2i(msg->message_tag) != tag)
	return -2;

    if (msg->message_type != RTS_ACKNOWLEDGE)
	return -3;

    return 0;
}




const char *
rts_status_to_string(int status)
{
    switch (status)
    {
      case   0: return "Operation Completed Successfully";
      case   1: return "Operation Started Successfully";
      case   2: return "Incomplete Status";
      case   3: return "Invalid Field";
      case   4: return "Duplicate Tag";
      case   5: return "Invalid Tag";
      case   6: return "Insufficent Privilege";
      case   7: return "Invalid Password";
      case   8: return "16-bit SCSI ID specified with 8-bit bus";
      case   9: return "16-bit Sync. Rate specified with 8-bit bus";
      case  10: return "Shutdown not initiated";
      case  11: return "Cache not flushed";
      case  12: return "8-bit module selected as host module";
      case  13: return "Invalid sequence number";
      case  14: return "Duplicate redundancy groups";
      case  15: return "No module installed on selected channel";
      case  16: return "Failed module on selected channel";
      case  17: return "Non-disk module on selected channel";
      case  18: return "Drive not empty";
      case  19: return "RTS internal inconsistency";
      case  20: return "System locked due to critical environmental data";
      case  21: return "No members to rebuild on selected RAID set";
      case  22: return "RAID set not degraded";
      case  23: return "Drive too small";
      case  24: return "Invalid capacity (capacity = 0)";
      case  25: return "Selected capacity too big";
      case  26: return "Chunk size too large";
      case  27: return "Too few drives";
      case  28: return "Too many drives";
      case  29: return "- Unused -";
      case  30: return "Odd number of drives not allowed (RAID level 0+1)";
      case  31: return "RAID set number already in use";
      case  32: return "Redundancy group number already in use";
      case  33: return "Create aborted";
      case  34: return "Rebuild aborted";
      case  35: return "Drive not online";
      case  36: return "Non-host module on selected channel";
      case  37: return "Non-wide module on selected channel";
      case  38: return "Redundancy group belongs to other controller";
      case  39: return "RAID set belongs to other controller";
      case  40: return "Warning level > Severe level temperature";
      case  41: return "Invalid message size";
      case  42: return "SCSI ID is reserved for controller";
      case  43: return "SCSI ID is reserved for failover";
      case  44: return "Session unavailable";
      case  45: return "Not a SCSI port message";
      case  46: return "Not a RS232 port message";
      case  47: return "Non-Ultra module";
      case  48: return "Exceed format drive limit";
      case  49: return "RAID set does not exist";
      case  50: return "Host not active on this controller";
      case  51: return "Attempting Auto-Rebind";
      case  52: return "Invalid channel configuration";
      case  53: return "Controller not part of redundant pair";

      case 128: return "Online member can not be selected";
      case 129: return "Creating member can not be selected";
      case 130: return "Rebuilding member can not be selected";
      case 131: return "Spare being added can not be selected";
      case 132: return "Formatting member can not be selected";
      case 133: return "Drive not detected";
      case 134: return "Non-disk device";
      case 135: return "Drive failed Inquiry";
      case 136: return "Drive failed Test Unit Ready";
      case 137: return "Drive failed Wide Negotiation";
      case 138: return "Drive failed Sync. Negotiation";
      case 139: return "Drive failed Unit Attention";
      case 140: return "Drive failed Read Capacity";
      case 141: return "Drive failed Spin Up";
      case 142: return "Drive failed Spin Down";
      case 143: return "Drive failed Read";
      case 144: return "Drive failed Read Long";
      case 145: return "Drive failed Write";
      case 146: return "Drive failed Home block Write";
      case 147: return "Drive failed Bus Device Reset";
      case 148: return "Invalid Read Long Sense Data";
      case 149: return "ECC size too large";
      case 150: return "Incorrect Block Size";
      case 151: return "Rebuilding drive failed non-rebuild operation";
      case 152: return "Rebuild failed due to RAID set going offline";
      case 153: return "Disk drive belongs to other controller";
      case 154: return "Raid set belongs to other controller";
      case 155: return "Create failed due to other controller";
      case 156: return "Failed set block capacity";
      case 157: return "Format failed";
      case 158: return "Failed Mode Sense";
      case 159: return "Failed Mode Select";

      case 254: return "SCSI RTS wrong message received";
		    
      default: return "Unknown error";
    }
}


void
rts_clearerr(RTS *rp)
{
    rp->unix_error_code = 0;
    memset(&rp->ack, 0, sizeof(rp->ack));
}


const char *
rts_strpstatus(int port_status)
{
    static char buf[256];

    
    switch (port_status)
    {
      case 0:
	return "Good";
		
      case 1:
	return "Reselection Tag Error - Invalid Tagged Queue Command Relselection";
		
      case 2:
	return "Check Condition while performing Request Sense";
		
      case 3:
	return "Received Illegal SCSI Status";
		
      case 4:
	return "Received Unsupported Message-In Byte";
		
      case 5:
	return "Selection Time-out";
		
      case 6:
	return "Illegal Disconnect";
		
      case 7:
	return "Target did not disconnect when expected";
		
      case 8:
	return "Detected SCSI Bus Reset";
		
      case 9:
	return "Disconnect Time-out";
		
      case 10:
	return "Illegal Reselection";
		
      case 11:
	return "Target rejected a non-rejectable Message-Out byte";
		
      case 12:
	return "SCSI Bus Time-out #1 - Timed out a Phase Change on the SCSI Bus";
		
      case 13:
	return "SCSI Bus Time-out #2 - The Target has been connected to the SCSI Bus for 30 seconds";
		
      case 14:
	return "SCSI Chip detected an Internal Gross Error";
		
      case 15:
	return "SCSI Chip detected a Parity Error on the SCSI Bus";
		
      case 16:
	return "SCSI Chip received an Illegal Command";
		
      case 17:
	return "Target has been taken Offline";
		
      case 18:
	return "Invalid Ignore Wide Residue Message was received";
		
      case 19:
	return "Target always busy";
		
      case 20:
	return "Odd byte in FIFO (wide only";
		
      case 21:
	return "Bad Opcode";
		
      case 22:
	return "Bad Message-Out";
		
      case 23:
	return "Failed Read Long";
		
      case 24:
	return "ECC Size Too Larget";
		
      case 25:
	return "Invalid Read Long Sense Data";
		
      case 26:
	return "Failed Device Reset";
		
      case 27:
	return "Failed Synchronous Negation";
		
      case 28:
	return "Port Failed Setup";
		
      case 29:
	return "Expected Simple Tag";
		
      case 30:
	return "Expected Message-In after Simple Tag";
		
      case 31:
	return "Reselection with Invalid Tag Number";
		
      case 32:
	return "Invalid Reselection ID";
		
      case 33:
	return "Target failed to abort the reselection";
		
      case 34:
	return "Target failed to go to Message-Out after Unit Attention Asserted";
		
      case 35:
	return "Request Sense command received BUSY Status";
		
      case 36:
	return "Request Sense command received Unsupported Status";
		
      case 37:
	return "Unexpected Disconnect after Message-Out";
		
      case 38:
	return "Unexpected Disconnect during a Data Phase";
		
      case 39:
	return "Internally generated Request Sense had a Disconnect Timeout";
		
      case 40:
	return "Internally generated Request Sense detected SCSI Bus Reset";
		
      case 41:
	return "REQ Overflow detected during a Data Phase";
		
      case 42:
	return "Data Out phase detected when Data In phase expected";
		
      case 43:
	return "Data In phase detected when Data Out phase expected";
		
      case 44:
	return "Block Type command disconnected on Non-block boundary ";
		
      case 45:
	return "Not enough data transferred during a Data phase";
		
      default:
	sprintf(buf, "%d", port_status);
        return buf;
    }
}


const char *
rts_strerror(RTS *rp)
{
    static char buf[1024];
    char *cp;

    
    if (rp->unix_error_code)
    {
	strcpy(buf, "Unix: ");
	strcat(buf, strerror(rp->unix_error_code));
	return buf;
    }

    strcpy(buf, "RTS: ");
    strcat(buf, rts_status_to_string(rp->ack.s.status));
    cp = buf + strlen(buf);

    if (rp->ack.s.status < 128)
    {
	switch (rp->ack.s.status)
	{
	  case 3:
	    sprintf(cp, " (at byte: %d)", rp->ack.s.additional_data);
	    break;
	    
	  case 14:
	    sprintf(cp, " (duplicated group: %d)", rp->ack.s.additional_data);
	    break;
	    
	  case 15:
	    sprintf(cp, " (channel: %d)", rp->ack.s.additional_data);
	    break;
	    
	  case 16:
	    sprintf(cp, " (channel: %d)", rp->ack.s.additional_data);
	    break;
	    
	  case 17:
	    sprintf(cp, " (channel: %d)", rp->ack.s.additional_data);
	    break;
	}
    }
    else
    {
	sprintf(cp, "(channel: %d, scsi: id %d",
		rp->ack.l.channel,
		rp->ack.l.scsi_id);

	cp += strlen(cp);
	
	if (rp->ack.l.scsi_status != 255)
	{
	    strcpy(cp, ", scsi status: ");
	    switch (rp->ack.l.scsi_status)
	    {
	      case 0:
		strcat(cp, "Good");
		break;

	      case 2:
		strcat(cp, "Check Condition");
		break;

	      default:
		sprintf(cp, "#%d", rp->ack.l.scsi_status);
	    }
	    cp += strlen(cp);
	}

	if (rp->ack.l.port_status != 255)
	{
	    strcpy(cp, ", port status: ");
	    strcat(cp, rts_strpstatus(rp->ack.l.port_status));
	    cp += strlen(cp);
	}
	
	strcpy(cp, ")");
    }
    
    return buf;
}


void
rts_perror(FILE *fp,
	   const char *prompt,
	   RTS *rp)	   
{
    if (prompt)
	fprintf(fp, "%s: %s\n", prompt, rts_strerror(rp));
    else
	fprintf(fp, "%s\n", rts_strerror(rp));
}


RTS *
rts_open(const char *device)
{
    RTS *rp;
    struct rts_static_system_information ssi;
    
    
    rp = malloc(sizeof(*rp));
    if (rp == NULL)
	return NULL;
    
    memset(rp, 0, sizeof(*rp));
    
    rp->fd = open(device, O_RDONLY);
    if (rp->fd < 0)
    {
	free(rp);
	return NULL;
    }

    rp->lockfile = strdup(PATH_LOCKFILE);

    if (rts_read_static_system_information(rp, &ssi) == 0)
	rp->version = ssi.rts_version;
    else
	rp->version = -1;
    
    return rp;
}


int
rts_close(RTS *rp)
{
    int code;

    while (rts_unlock(rp) == 0)
	;
    code = close(rp->fd);
    free(rp->lockfile);
    free(rp);

    return code;
}


int
rts_lock(RTS *rp)
{
    if (debug)
	fprintf(stderr, "rts_lock()\n");
    
    rp->lock++;
    if (rp->lock == 1)
    {
	int fd, i, pid, len;

      Again:	
	for (i = 0; i < MAX_LOCK_ATTEMPTS; i++)
	{
	    fd = open(rp->lockfile, O_WRONLY|O_CREAT|O_TRUNC|O_EXCL, 0600);
	    if (fd >= 0)
	    {
		pid = getpid();
		if (write(fd, &pid, sizeof(pid)) != sizeof(pid))
		{
		    rp->unix_error_code = errno;
		    unlink(rp->lockfile);
		    close(fd);
		    return -1;
		}
		close(fd);

		if (debug)
		    fprintf(stderr, "Created %s\n", rp->lockfile);
		break;
	    }

	    fd = open(rp->lockfile, O_RDONLY);
	    if (fd >= 0)
	    {
		len = read(fd, &pid, sizeof(pid));
		close(fd);
		if (len == sizeof(pid))
		{
		    if (kill(pid, 0) < 0 && errno == ESRCH)
		    {
			/* Stale lock - remove it */
			if (debug)
			    fprintf(stderr,
				    "rts_lock: Stale lock (was owned by pid %d)\n",
				    pid);
			unlink(rp->lockfile);
			goto Again;
		    }
		}
	    }

	    if (debug)
		fprintf(stderr, "rts_lock: Lock busy, sleeping 1 second\n");
	    sleep(1);
	}

	if (i >= MAX_LOCK_ATTEMPTS)
	{
	    if (debug)
		fprintf(stderr, "rts_lock: Lock busy: Failed");
	    rp->unix_error_code = errno;
	    return -1;
	}
    }
	
    return 0;
}


int
rts_unlock(RTS *rp)
{
    if (debug)
	fprintf(stderr, "rts_unlock()\n");
    
    if (rp->lock <= 0)
    {
	rp->unix_error_code = EINVAL;
	return -1;
    }
    
    rp->lock--;
    if (rp->lock == 0)
    {
	unlink(rp->lockfile);
	if (debug)
	    fprintf(stderr, "Unlinked %s\n", rp->lockfile);
    }
    
    return 0;
}



int
rts_flush_messages(RTS *rp)
{
    struct rts_message msg;
    int code;
    

    if (rts_lock(rp) < 0)
	return -1;
    
    do
    {
	memset(&msg, 0, sizeof(msg));
	
	code = scsi_command(rp->fd, SCMD_GDIAG, USCSI_READ|USCSI_SILENT, 
			    (void *) &msg, sizeof(msg), 1);

	if (code == 0)
	{
	    if (debug)
	    {
		if (memcmp(msg.signature, " STR", 4) != 0)
		    printf("Got message with invalid signature\n");
		
		printf("Message tag: %u\n", rts_q2i(msg.message_tag));
		printf("Message type: %d\n", msg.message_type);
		printf("Message type extender: %d\n",
		       msg.message_type_extender);
	    }
	}
    } while (code == 0);

    rts_unlock(rp);
    
    return 0;
}



int
rts_command(RTS *rp,
	    int command_type,
	    void *in,
	    int inlen)
{
    struct rts_message msg;
    unsigned long tag;
    int code;
    

    rts_clearerr(rp);
    tag = rts_init_message(&msg, RTS_COMMAND, command_type,
			   rp->ownership_token);

    if (in)
	memcpy(&msg.body.command, in, inlen);

    if (rts_lock(rp) < 0)
	return -1;
    
    code = scsi_command(rp->fd, SCMD_SDIAG, USCSI_WRITE,
			(void  *) &msg, 12+inlen, SCSI_TIMEOUT);
    if (code != 0)
    {
	rp->unix_error_code = errno;
	rts_unlock(rp);
	return -1;
    }

    memset(&msg, 0, sizeof(msg));
    
    code = scsi_command(rp->fd, SCMD_GDIAG, USCSI_READ, 
			(void *) &msg, sizeof(msg), SCSI_TIMEOUT);
    
    if (code != 0)
    {
	rp->unix_error_code = errno;
	rts_unlock(rp);
	return -1;
    }

    rts_unlock(rp);
    
    if (rts_valid_acknowledge(&msg, tag) < 0)
    {
	rp->ack.s.status = 254;
	return -1;
    }

    rp->ack = msg.body.acknowledge;
    return msg.body.acknowledge.s.status;
}




/*
** Generic RTS Read command
*/

int
rts_read(RTS *rp,
	 int command_type,
	 void *in,
	 int inlen,
	 void *out,
	 int outlen)
{
    struct rts_message msg;
    unsigned long tag;
    int code;

    
    rts_clearerr(rp);
    tag = rts_init_message(&msg, RTS_READ, command_type,
			   rp->ownership_token);			   

    if (in)
	memcpy(&msg.body.command, in, inlen);

    if (rts_lock(rp) < 0)
	return -1;
    
    code = scsi_command(rp->fd, SCMD_SDIAG, USCSI_WRITE,
			    (void  *) &msg, 12+inlen, SCSI_TIMEOUT);
    if (code != 0)
    {
	rp->unix_error_code = errno;
	rts_unlock(rp);
	return -1;
    }
    
    memset(&msg, 0, sizeof(msg));
    
    code = scsi_command(rp->fd, SCMD_GDIAG, USCSI_READ, 
			(void *) &msg, sizeof(msg), SCSI_TIMEOUT);

    if (code != 0)
    {
	rp->unix_error_code = errno;
	rts_unlock(rp);
	return -1;
    }

    rts_unlock(rp);
    
    if (rts_valid_response(&msg, tag, command_type) < 0)
    {
	rp->ack.s.status = 254;
	return -1;
    }

    memcpy(out, &msg.body.response, outlen);
    return 0;
}



/*
** Generic RTS Write command
*/

int
rts_write(RTS *rp,
	  int command_type,
	  void *in,
	  int inlen)
{
    struct rts_message msg;
    unsigned long tag;
    int code;

    
    rts_clearerr(rp);
    tag = rts_init_message(&msg, RTS_WRITE, command_type,
			   rp->ownership_token);			   

    if (in)
	memcpy(&msg.body.write, in, inlen);

    if (rts_lock(rp) < 0)
	return -1;
    
    code = scsi_command(rp->fd, SCMD_SDIAG, USCSI_WRITE,
			(void  *) &msg, 12+inlen, SCSI_TIMEOUT);
    if (code != 0)
    {
	rp->unix_error_code = errno;
	rts_unlock(rp);
	return -1;
    }
    
    memset(&msg, 0, sizeof(msg));
    
    code = scsi_command(rp->fd, SCMD_GDIAG, USCSI_READ, 
			(void *) &msg, sizeof(msg), SCSI_TIMEOUT);

    if (code != 0)
    {
	rp->unix_error_code = errno;
	rts_unlock(rp);
	return -1;
    }

    rts_unlock(rp);
    
    if (rts_valid_acknowledge(&msg, tag) < 0)
    {
	rp->ack.s.status = 254;
	return -1;
    }

    rp->ack = msg.body.acknowledge;
    
    return msg.body.acknowledge.s.status;
}




const char *
rts_raid2str(int level)
{
    static char buf[64];

    if (level == 6)
	return "0+1";
    else if (level == 7)
	return "JBOD";

    sprintf(buf, "%d", level);
    return buf;
}


const char *
rts_status2str(int status)
{
    static char buf[64];

    switch (status)
    {
      case 0:
	return "Unused";
      case 1:
	return "Non-Degraded";
      case 2:
	return "Degraded";
      case 3:
	return "Offline";
      case 4:
	return "Creating";
    }
    
    sprintf(buf, "%d", status);
    return buf;
}



