/*
** rts.h
**
** Structure and definitions for Viper SCSI Remote Terminal Services
**
** Copyright (c) 1998 Peter Eriksson <pen@lysator.liu.se>
*/

#ifndef VIPER_RTS_H
#define VIPER_RTS_H

#include <time.h>

/* The version of the RTS protocol that the include files define */
#define RTS_VERSION 9


/* RTS Message Types */
#define RTS_READ				0
#define RTS_WRITE 				1
#define RTS_COMMAND				2
#define RTS_RESPONSE				3
#define RTS_ACKNOWLEDGE				4

/* First some basic type definitions */
typedef unsigned char Byte;

typedef struct
{
    unsigned char b[2];
} Word;

typedef struct
{
    unsigned char b[4];
} Quad;


/* Date And Time */
struct rts_date_and_time
{
    Byte month;
    Byte date;
    Byte year;
    Byte hour;
    Byte minute;
    Byte second;
};



/*
** RTS Acknowledge Messages
*/
union rts_acknowledge_message
{
    struct
    {
	Byte status;
	Byte additional_data;
    } s;
    struct
    {
	Byte status;
	Byte channel;
	Byte scsi_id;
	Byte scsi_status;
	Byte port_status;
	Byte request_sense;
    } l;
};


typedef struct
{
    int fd;
    int ownership_token;
    int lock;
    char *lockfile;
    int unix_error_code;
    int version;
    union rts_acknowledge_message ack;
} RTS;


#include <rts/command.h>
#include <rts/event.h>
#include <rts/rdwr.h>
#include <rts/print.h>




/*
** The generic RTS Message structure
*/
struct rts_message
{
    Byte signature[4];
    Byte protocol_version;
    Byte ownership_token;
    Byte message_type;
    Byte message_type_extender;
    Quad message_tag;
    union
    {
	union rts_command_message 	command;
	union rts_write_message		write;
	union rts_read_message		read;
	union rts_response_message	response;
	union rts_acknowledge_message	acknowledge;
	char dummy[244];
    } body;
};



/*
** Generic RTS functions ----------------------------------------
*/

extern unsigned int
rts_q2i(Quad q);

extern unsigned int
rts_w2i(Word w);

extern Quad
rts_i2q(unsigned int i);

extern Word
rts_i2w(unsigned int i);


extern struct rts_date_and_time *
rts_localtime(time_t t,
	      struct rts_date_and_time *rtp);

extern time_t
rts_mktime(struct rts_date_and_time *rtp);

extern const char *
rts_raid2str(int level);

extern const char *
rts_status2str(int status);

extern const char *
rts_strpstatus(int port_status);

extern int
rts_valid_response(struct rts_message *msg,
		   unsigned long tag,
		   int message_type_extender);

extern int
rts_valid_acknowledge(struct rts_message *msg,
		      unsigned long tag);

extern unsigned long
rts_init_message(struct rts_message *msg,
		 int message_type,
		 int message_type_extender,
		 int ownership_token);

extern void
rts_set_debug(int flag);

extern RTS *
rts_open(const char *device);

extern int
rts_close(RTS *rp);

extern int
rts_lock(RTS *rp);

extern int
rts_unlock(RTS *rp);

extern int
rts_command(RTS *rp,
	    int command_type,
	    void *in,
	    int inlen);

extern int
rts_read(RTS *rp,
	 int command_type,
	 void *in,
	 int inlen,
	 void *out,
	 int outlen);

extern int
rts_write(RTS *rp,
	  int command_type,
	  void *in,
	  int inlen);


extern const char *
rts_strerror(RTS *rp);

extern void
rts_perror(FILE *fp,
	   const char *prompt,
	   RTS *rp);

extern void
rts_clearerr(RTS *rp);

extern int
rts_flush_messages(RTS *rp);

extern const char *
rts_status_to_string(int status);


#endif
